"""Tests for CyclesClient and AsyncCyclesClient."""


import httpx
import pytest

from runcycles.client import AsyncCyclesClient, CyclesClient
from runcycles.config import CyclesConfig
from runcycles.models import (
    Action,
    Amount,
    CommitRequest,
    DecisionRequest,
    EventCreateRequest,
    ReleaseRequest,
    ReservationCreateRequest,
    ReservationExtendRequest,
    Subject,
    Unit,
)


@pytest.fixture
def config() -> CyclesConfig:
    return CyclesConfig(base_url="http://localhost:7878", api_key="test-key")


class TestCyclesClientSync:
    def test_create_reservation_success(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW", "reservation_id": "res_123"},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.create_reservation(ReservationCreateRequest(
                idempotency_key="req-001",
                subject=Subject(tenant="acme"),
                action=Action(kind="llm.completion", name="gpt-4"),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=1000),
            ))

        assert response.is_success
        assert response.get_body_attribute("decision") == "ALLOW"
        assert response.get_body_attribute("reservation_id") == "res_123"

    def test_create_reservation_denied(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"error": "BUDGET_EXCEEDED", "message": "Insufficient budget"},
            status_code=409,
        )

        with CyclesClient(config) as client:
            response = client.create_reservation({"idempotency_key": "req-002", "subject": {"tenant": "acme"}})

        assert not response.is_success
        assert response.is_client_error
        assert response.status == 409

    def test_commit_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_123/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 800}},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.commit_reservation("res_123", CommitRequest(
                idempotency_key="commit-001",
                actual=Amount(unit=Unit.USD_MICROCENTS, amount=800),
            ))

        assert response.is_success
        assert response.get_body_attribute("status") == "COMMITTED"

    def test_release_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_123/release",
            json={"status": "RELEASED"},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.release_reservation("res_123", ReleaseRequest(
                idempotency_key="rel-001",
                reason="cancelled",
            ))

        assert response.is_success

    def test_extend_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_123/extend",
            json={"status": "ACTIVE", "expires_at_ms": 9999999},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.extend_reservation("res_123", ReservationExtendRequest(
                idempotency_key="ext-001",
                extend_by_ms=60000,
            ))

        assert response.is_success

    def test_decide(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/decide",
            json={"decision": "ALLOW"},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.decide(DecisionRequest(
                idempotency_key="dec-001",
                subject=Subject(tenant="acme"),
                action=Action(kind="llm.completion", name="gpt-4"),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=1000),
            ))

        assert response.is_success

    def test_get_balances(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="GET",
            url="http://localhost:7878/v1/balances?tenant=acme",
            json={"balances": []},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.get_balances(tenant="acme")

        assert response.is_success

    def test_list_reservations(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="GET",
            url="http://localhost:7878/v1/reservations?tenant=acme",
            json={"reservations": [], "has_more": False},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.list_reservations(tenant="acme")

        assert response.is_success

    def test_get_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="GET",
            url="http://localhost:7878/v1/reservations/res_123",
            json={"reservation_id": "res_123", "status": "ACTIVE"},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.get_reservation("res_123")

        assert response.is_success

    def test_create_event(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/events",
            json={"status": "APPLIED", "event_id": "evt_123"},
            status_code=201,
        )

        with CyclesClient(config) as client:
            response = client.create_event(EventCreateRequest(
                idempotency_key="evt-001",
                subject=Subject(tenant="acme"),
                action=Action(kind="llm.completion", name="gpt-4"),
                actual=Amount(unit=Unit.USD_MICROCENTS, amount=500),
            ))

        assert response.is_success

    def test_transport_error(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_exception(httpx.ConnectError("Connection refused"))

        with CyclesClient(config) as client:
            response = client.create_reservation({"idempotency_key": "req-err"})

        assert response.is_transport_error
        assert response.status == -1

    def test_idempotency_header_set(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW"},
            status_code=200,
        )

        with CyclesClient(config) as client:
            client.create_reservation({"idempotency_key": "my-key", "subject": {"tenant": "acme"}})

        request = httpx_mock.get_request()
        assert request is not None
        assert request.headers.get("X-Idempotency-Key") == "my-key"
        assert request.headers.get("X-Cycles-API-Key") == "test-key"

    def test_get_balances_requires_filter(self, config: CyclesConfig) -> None:
        with CyclesClient(config) as client:
            with pytest.raises(ValueError, match="at least one subject filter"):
                client.get_balances()

    def test_get_balances_non_subject_params_rejected(self, config: CyclesConfig) -> None:
        with CyclesClient(config) as client:
            with pytest.raises(ValueError, match="at least one subject filter"):
                client.get_balances(limit="10")

    def test_response_headers_extracted(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW"},
            status_code=200,
            headers={"X-Request-Id": "req-abc", "X-RateLimit-Remaining": "42"},
        )

        with CyclesClient(config) as client:
            response = client.create_reservation({"idempotency_key": "hdr-test", "subject": {"tenant": "acme"}})

        assert response.request_id == "req-abc"
        assert response.rate_limit_remaining == 42

    def test_dict_body(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW"},
            status_code=200,
        )

        with CyclesClient(config) as client:
            response = client.create_reservation({
                "idempotency_key": "raw-dict",
                "subject": {"tenant": "acme"},
                "estimate": {"unit": "USD_MICROCENTS", "amount": 1000},
            })

        assert response.is_success


class TestCyclesClientSyncEdgeCases:
    def test_unsupported_body_type_raises(self, config: CyclesConfig) -> None:
        with CyclesClient(config) as client:
            with pytest.raises(TypeError, match="Unsupported body type"):
                client.create_reservation("not-a-dict-or-model")  # type: ignore[arg-type]

    def test_get_transport_error(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_exception(httpx.ConnectError("Connection refused"))

        with CyclesClient(config) as client:
            response = client.list_reservations(tenant="acme")

        assert response.is_transport_error
        assert response.status == -1

    def test_json_parse_failure(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            content=b"not-json",
            status_code=200,
            headers={"content-type": "text/plain"},
        )

        with CyclesClient(config) as client:
            response = client.create_reservation({"idempotency_key": "bad-json"})

        assert response.is_success
        assert response.body == {}

    def test_json_parse_failure_error_response(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            content=b"internal error",
            status_code=500,
            headers={"content-type": "text/plain"},
        )

        with CyclesClient(config) as client:
            response = client.create_reservation({"idempotency_key": "bad-json-err"})

        assert response.is_server_error


@pytest.mark.asyncio
class TestAsyncCyclesClient:
    async def test_create_reservation_success(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW", "reservation_id": "res_456"},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.create_reservation(ReservationCreateRequest(
                idempotency_key="async-001",
                subject=Subject(tenant="acme"),
                action=Action(kind="llm.completion", name="gpt-4"),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=1000),
            ))

        assert response.is_success
        assert response.get_body_attribute("reservation_id") == "res_456"

    async def test_commit_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_a1/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 800}},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.commit_reservation("res_a1", CommitRequest(
                idempotency_key="ac-001",
                actual=Amount(unit=Unit.USD_MICROCENTS, amount=800),
            ))

        assert response.is_success
        assert response.get_body_attribute("status") == "COMMITTED"

    async def test_release_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_a1/release",
            json={"status": "RELEASED", "released": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.release_reservation("res_a1", ReleaseRequest(
                idempotency_key="ar-001", reason="done",
            ))

        assert response.is_success

    async def test_extend_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_a1/extend",
            json={"status": "ACTIVE", "expires_at_ms": 9999999},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.extend_reservation("res_a1", ReservationExtendRequest(
                idempotency_key="ae-001", extend_by_ms=60000,
            ))

        assert response.is_success

    async def test_decide(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/decide",
            json={"decision": "ALLOW"},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.decide(DecisionRequest(
                idempotency_key="ad-001",
                subject=Subject(tenant="acme"),
                action=Action(kind="llm.completion", name="gpt-4"),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=1000),
            ))

        assert response.is_success

    async def test_list_reservations(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="GET",
            url="http://localhost:7878/v1/reservations?tenant=acme",
            json={"reservations": [], "has_more": False},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.list_reservations(tenant="acme")

        assert response.is_success

    async def test_get_reservation(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="GET",
            url="http://localhost:7878/v1/reservations/res_a1",
            json={"reservation_id": "res_a1", "status": "ACTIVE"},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.get_reservation("res_a1")

        assert response.is_success

    async def test_get_balances(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="GET",
            url="http://localhost:7878/v1/balances?tenant=acme",
            json={"balances": []},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.get_balances(tenant="acme")

        assert response.is_success

    async def test_get_balances_requires_filter(self, config: CyclesConfig) -> None:
        async with AsyncCyclesClient(config) as client:
            with pytest.raises(ValueError, match="at least one subject filter"):
                await client.get_balances()

    async def test_create_event(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/events",
            json={"status": "APPLIED", "event_id": "evt_a1"},
            status_code=201,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.create_event(EventCreateRequest(
                idempotency_key="aevt-001",
                subject=Subject(tenant="acme"),
                action=Action(kind="llm.completion", name="gpt-4"),
                actual=Amount(unit=Unit.USD_MICROCENTS, amount=500),
            ))

        assert response.is_success

    async def test_transport_error(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_exception(httpx.ConnectError("Connection refused"))

        async with AsyncCyclesClient(config) as client:
            response = await client.create_reservation({"idempotency_key": "async-err"})

        assert response.is_transport_error

    async def test_get_transport_error(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_exception(httpx.ConnectError("Connection refused"))

        async with AsyncCyclesClient(config) as client:
            response = await client.list_reservations(tenant="acme")

        assert response.is_transport_error

    async def test_json_parse_failure(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            content=b"not-json",
            status_code=200,
            headers={"content-type": "text/plain"},
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.create_reservation({"idempotency_key": "bad"})

        assert response.is_success
        assert response.body == {}

    async def test_idempotency_header_set(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW"},
            status_code=200,
        )

        async with AsyncCyclesClient(config) as client:
            await client.create_reservation({"idempotency_key": "async-key", "subject": {"tenant": "acme"}})

        request = httpx_mock.get_request()
        assert request is not None
        assert request.headers.get("X-Idempotency-Key") == "async-key"

    async def test_error_response_with_body(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"error": "BUDGET_EXCEEDED", "message": "No budget"},
            status_code=409,
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.create_reservation({"idempotency_key": "err-async"})

        assert response.is_client_error
        assert response.status == 409

    async def test_error_response_no_body(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            content=b"not-json",
            status_code=500,
            headers={"content-type": "text/plain"},
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.create_reservation({"idempotency_key": "err-no-body"})

        assert response.is_server_error

    async def test_response_headers_extracted(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"decision": "ALLOW"},
            status_code=200,
            headers={"X-Request-Id": "async-req-id", "X-Cycles-Tenant": "acme"},
        )

        async with AsyncCyclesClient(config) as client:
            response = await client.create_reservation({"idempotency_key": "hdr-async"})

        assert response.request_id == "async-req-id"
        assert response.cycles_tenant == "acme"



"""Input validation utilities."""

from __future__ import annotations

from runcycles.models import Subject


def validate_subject(subject: Subject | None) -> None:
    """Validate that a subject has at least one standard field."""
    if subject is not None and not subject.has_at_least_one_standard_field():
        raise ValueError(
            "Subject must have at least one standard field"
            " (tenant, workspace, app, workflow, agent, or toolset)"
        )


def validate_reservation_id(reservation_id: str | None) -> None:
    """Validate that a reservation ID is non-empty."""
    if not reservation_id:
        raise ValueError("reservation_id is required and must be non-empty")


def validate_non_negative(value: int, name: str) -> None:
    """Validate that a value is non-negative (>= 0), per spec Amount minimum: 0."""
    if value < 0:
        raise ValueError(f"{name} must be non-negative, got {value}")


def validate_ttl_ms(ttl_ms: int) -> None:
    """Validate TTL is within allowed range (1s to 24h)."""
    if ttl_ms < 1000 or ttl_ms > 86_400_000:
        raise ValueError(f"ttl_ms must be between 1000 and 86400000, got {ttl_ms}")


def validate_extend_by_ms(extend_by_ms: int) -> None:
    """Validate extend_by_ms is within allowed range (1ms to 24h)."""
    if extend_by_ms < 1 or extend_by_ms > 86_400_000:
        raise ValueError(f"extend_by_ms must be between 1 and 86400000, got {extend_by_ms}")


def validate_grace_period_ms(grace_period_ms: int | None) -> None:
    """Validate grace period is within allowed range (0 to 60s)."""
    if grace_period_ms is not None and (grace_period_ms < 0 or grace_period_ms > 60_000):
        raise ValueError(f"grace_period_ms must be between 0 and 60000, got {grace_period_ms}")



"""Tests for the commit retry engine."""

from __future__ import annotations

import asyncio
from unittest.mock import AsyncMock, MagicMock

import pytest

from runcycles.config import CyclesConfig
from runcycles.response import CyclesResponse
from runcycles.retry import AsyncCommitRetryEngine, CommitRetryEngine


@pytest.fixture
def config() -> CyclesConfig:
    return CyclesConfig(
        base_url="http://localhost:7878",
        api_key="test-key",
        retry_enabled=True,
        retry_max_attempts=3,
        retry_initial_delay=0.01,  # fast for tests
        retry_multiplier=1.0,
        retry_max_delay=0.05,
    )


@pytest.fixture
def disabled_config() -> CyclesConfig:
    return CyclesConfig(
        base_url="http://localhost:7878",
        api_key="test-key",
        retry_enabled=False,
    )


class TestCommitRetryEngine:
    def test_disabled_does_not_retry(self, disabled_config: CyclesConfig) -> None:
        engine = CommitRetryEngine(disabled_config)
        mock_client = MagicMock()
        engine.set_client(mock_client)
        engine.schedule("rsv_1", {"idempotency_key": "k1", "actual": {"unit": "USD_MICROCENTS", "amount": 100}})
        mock_client.commit_reservation.assert_not_called()

    def test_retries_until_success(self, config: CyclesConfig) -> None:
        engine = CommitRetryEngine(config)
        mock_client = MagicMock()
        # First call fails with 500, second succeeds
        mock_client.commit_reservation.side_effect = [
            CyclesResponse.http_error(500, "Server error"),
            CyclesResponse.success(200, {"status": "COMMITTED"}),
        ]
        engine.set_client(mock_client)

        # Run _retry_loop directly (synchronous, avoids thread timing issues)
        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == 2

    def test_stops_on_client_error(self, config: CyclesConfig) -> None:
        engine = CommitRetryEngine(config)
        mock_client = MagicMock()
        # 409 is a client error — should not retry
        mock_client.commit_reservation.return_value = CyclesResponse.http_error(409, "Already finalized")
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == 1

    def test_exhausts_retries(self, config: CyclesConfig) -> None:
        engine = CommitRetryEngine(config)
        mock_client = MagicMock()
        # Always returns 500
        mock_client.commit_reservation.return_value = CyclesResponse.http_error(500, "Server error")
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == config.retry_max_attempts

    def test_handles_exception_during_retry(self, config: CyclesConfig) -> None:
        engine = CommitRetryEngine(config)
        mock_client = MagicMock()
        # First call throws, second succeeds
        mock_client.commit_reservation.side_effect = [
            ConnectionError("network down"),
            CyclesResponse.success(200, {"status": "COMMITTED"}),
        ]
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == 2

    def test_no_client_set(self, config: CyclesConfig) -> None:
        engine = CommitRetryEngine(config)
        # Don't set client — should bail out
        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        engine._retry_loop(pending)  # should not raise


@pytest.mark.asyncio
class TestAsyncCommitRetryEngine:
    async def test_disabled_does_not_retry(self, disabled_config: CyclesConfig) -> None:
        engine = AsyncCommitRetryEngine(disabled_config)
        mock_client = AsyncMock()
        engine.set_client(mock_client)
        engine.schedule("rsv_1", {"idempotency_key": "k1"})
        mock_client.commit_reservation.assert_not_called()

    async def test_retries_until_success(self, config: CyclesConfig) -> None:
        engine = AsyncCommitRetryEngine(config)
        mock_client = AsyncMock()
        mock_client.commit_reservation.side_effect = [
            CyclesResponse.http_error(500, "Server error"),
            CyclesResponse.success(200, {"status": "COMMITTED"}),
        ]
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        await engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == 2

    async def test_stops_on_client_error(self, config: CyclesConfig) -> None:
        engine = AsyncCommitRetryEngine(config)
        mock_client = AsyncMock()
        mock_client.commit_reservation.return_value = CyclesResponse.http_error(409, "Finalized")
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        await engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == 1

    async def test_exhausts_retries(self, config: CyclesConfig) -> None:
        engine = AsyncCommitRetryEngine(config)
        mock_client = AsyncMock()
        mock_client.commit_reservation.return_value = CyclesResponse.http_error(500, "Error")
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        await engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == config.retry_max_attempts

    async def test_handles_exception_during_retry(self, config: CyclesConfig) -> None:
        engine = AsyncCommitRetryEngine(config)
        mock_client = AsyncMock()
        mock_client.commit_reservation.side_effect = [
            ConnectionError("network down"),
            CyclesResponse.success(200, {"status": "COMMITTED"}),
        ]
        engine.set_client(mock_client)

        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        await engine._retry_loop(pending)

        assert mock_client.commit_reservation.call_count == 2

    async def test_no_client_set(self, config: CyclesConfig) -> None:
        engine = AsyncCommitRetryEngine(config)
        from runcycles.retry import _PendingCommit
        pending = _PendingCommit(reservation_id="rsv_1", commit_body={"idempotency_key": "k1"})
        await engine._retry_loop(pending)  # should not raise

    async def test_schedule_no_event_loop(self, config: CyclesConfig) -> None:
        """Schedule outside an event loop should log error, not crash."""
        engine = AsyncCommitRetryEngine(config)
        mock_client = AsyncMock()
        engine.set_client(mock_client)
        # Can't easily test outside event loop from within async test,
        # but we can verify schedule works inside one
        engine.schedule("rsv_1", {"idempotency_key": "k1"})
        # Let the scheduled task actually run
        await asyncio.sleep(0.1)


class TestCommitRetryEngineSchedule:
    def test_schedule_creates_thread(self, config: CyclesConfig) -> None:
        engine = CommitRetryEngine(config)
        mock_client = MagicMock()
        # Return success immediately so the thread finishes quickly
        mock_client.commit_reservation.return_value = CyclesResponse.success(200, {"status": "COMMITTED"})
        engine.set_client(mock_client)

        engine.schedule("rsv_1", {"idempotency_key": "k1"})
        # Give the thread time to run
        import time
        time.sleep(0.1)
        assert mock_client.commit_reservation.call_count >= 1


class TestAsyncCommitRetryEngineScheduleNoLoop:
    def test_schedule_without_event_loop(self, config: CyclesConfig) -> None:
        """Schedule outside an event loop should log error, not crash."""
        engine = AsyncCommitRetryEngine(config)
        mock_client = MagicMock()
        engine.set_client(mock_client)
        # This should not raise even without an event loop
        engine.schedule("rsv_1", {"idempotency_key": "k1"})



"""Tests for the @cycles decorator."""

import pytest

from runcycles.client import AsyncCyclesClient, CyclesClient
from runcycles.config import CyclesConfig
from runcycles.context import get_cycles_context
from runcycles.decorator import cycles, set_default_client, set_default_config
from runcycles.exceptions import CyclesProtocolError


@pytest.fixture
def config() -> CyclesConfig:
    return CyclesConfig(base_url="http://localhost:7878", api_key="test-key", tenant="acme")


class TestCyclesDecoratorSync:
    def test_basic_lifecycle(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        # Mock reservation creation
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW",
                "reservation_id": "res_dec_1",
                "expires_at_ms": 9999999999,
                "affected_scopes": ["tenant:acme"],
                "scope_path": "tenant:acme",
                "reserved": {"unit": "USD_MICROCENTS", "amount": 1000},
            },
            status_code=200,
        )
        # Mock commit
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_dec_1/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        client = CyclesClient(config)

        @cycles(estimate=1000, client=client)
        def my_func(x: int) -> str:
            ctx = get_cycles_context()
            assert ctx is not None
            assert ctx.reservation_id == "res_dec_1"
            return f"result-{x}"

        result = my_func(42)
        assert result == "result-42"
        client.close()

    def test_callable_estimate_and_actual(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_dec_2",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_dec_2/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 500}},
            status_code=200,
        )

        client = CyclesClient(config)

        @cycles(
            estimate=lambda x: x * 10,
            actual=lambda result: len(result) * 5,
            client=client,
        )
        def compute(x: int) -> str:
            return "hello"

        result = compute(100)
        assert result == "hello"
        client.close()

    def test_denied_raises(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={"error": "BUDGET_EXCEEDED", "message": "No budget"},
            status_code=409,
        )

        client = CyclesClient(config)

        @cycles(estimate=1000, client=client)
        def guarded() -> str:
            return "should not run"

        with pytest.raises(CyclesProtocolError):
            guarded()
        client.close()

    def test_function_exception_triggers_release(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_dec_3",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_dec_3/release",
            json={"status": "RELEASED", "released": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        client = CyclesClient(config)

        @cycles(estimate=1000, client=client)
        def failing_func() -> str:
            raise RuntimeError("boom")

        with pytest.raises(RuntimeError, match="boom"):
            failing_func()
        client.close()

    def test_context_cleared_after_call(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_dec_4",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_dec_4/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        client = CyclesClient(config)

        @cycles(estimate=1000, client=client)
        def func() -> str:
            assert get_cycles_context() is not None
            return "ok"

        func()
        assert get_cycles_context() is None
        client.close()


@pytest.mark.asyncio
class TestCyclesDecoratorAsync:
    async def test_basic_async_lifecycle(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_async_1",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_async_1/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        client = AsyncCyclesClient(config)

        @cycles(estimate=1000, client=client)
        async def async_func(x: int) -> str:
            ctx = get_cycles_context()
            assert ctx is not None
            return f"async-{x}"

        result = await async_func(42)
        assert result == "async-42"
        await client.aclose()


class TestDefaultClientConfig:
    def test_set_default_client(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_def_1",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_def_1/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        client = CyclesClient(config)
        set_default_client(client)

        @cycles(estimate=1000)
        def func() -> str:
            return "ok"

        result = func()
        assert result == "ok"
        client.close()

    def test_set_default_config_creates_client_lazily(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_lazy_1",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_lazy_1/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        set_default_config(config)

        @cycles(estimate=1000)
        def func() -> str:
            return "lazy"

        result = func()
        assert result == "lazy"

    def test_no_client_raises(self) -> None:
        @cycles(estimate=1000)
        def func() -> str:
            return "nope"

        with pytest.raises(ValueError, match="No Cycles client available"):
            func()

    def test_sync_func_with_async_client_raises(self, config: CyclesConfig) -> None:
        async_client = AsyncCyclesClient(config)
        set_default_client(async_client)

        @cycles(estimate=1000)
        def func() -> str:
            return "nope"

        with pytest.raises(TypeError, match="Sync function requires a CyclesClient"):
            func()

    @pytest.mark.asyncio
    async def test_set_default_config_creates_async_client_lazily(self, config: CyclesConfig, httpx_mock) -> None:  # type: ignore[no-untyped-def]
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations",
            json={
                "decision": "ALLOW", "reservation_id": "res_lazy_a1",
                "expires_at_ms": 9999999999, "affected_scopes": ["tenant:acme"],
            },
            status_code=200,
        )
        httpx_mock.add_response(
            method="POST",
            url="http://localhost:7878/v1/reservations/res_lazy_a1/commit",
            json={"status": "COMMITTED", "charged": {"unit": "USD_MICROCENTS", "amount": 1000}},
            status_code=200,
        )

        set_default_config(config)

        @cycles(estimate=1000)
        async def func() -> str:
            return "async-lazy"

        result = await func()
        assert result == "async-lazy"

    @pytest.mark.asyncio
    async def test_async_func_with_sync_client_raises(self, config: CyclesConfig) -> None:
        sync_client = CyclesClient(config)
        set_default_client(sync_client)

        @cycles(estimate=1000)
        async def func() -> str:
            return "nope"

        with pytest.raises(TypeError, match="Async function requires an AsyncCyclesClient"):
            await func()



"""Integrating Cycles with LangChain.

Demonstrates a custom callback handler that creates budget reservations
for every LLM call in a LangChain chain or agent.

Requirements:
    pip install runcycles langchain langchain-openai

Environment variables:
    CYCLES_BASE_URL  - Cycles server URL (default: http://localhost:7878)
    CYCLES_API_KEY   - Cycles API key
    CYCLES_TENANT    - Tenant identifier
    OPENAI_API_KEY   - OpenAI API key
"""

from __future__ import annotations

import os
import uuid
from typing import Any

from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.messages import HumanMessage
from langchain_core.outputs import LLMResult
from langchain_openai import ChatOpenAI

from runcycles import (
    Action,
    Amount,
    BudgetExceededError,
    CommitRequest,
    CyclesClient,
    CyclesConfig,
    CyclesMetrics,
    CyclesProtocolError,
    ReleaseRequest,
    ReservationCreateRequest,
    Subject,
    Unit,
)

# Pricing in USD microcents
PRICE_PER_INPUT_TOKEN = 250
PRICE_PER_OUTPUT_TOKEN = 1_000


# ---------------------------------------------------------------------------
# 1. Custom Callback Handler
# ---------------------------------------------------------------------------
class CyclesBudgetHandler(BaseCallbackHandler):
    """LangChain callback handler that wraps each LLM call with a Cycles reservation.

    Usage:
        handler = CyclesBudgetHandler(client, subject=Subject(tenant="acme"))
        llm = ChatOpenAI(callbacks=[handler])
    """

    def __init__(
        self,
        client: CyclesClient,
        subject: Subject,
        estimate_amount: int = 2_000_000,
        action_kind: str = "llm.completion",
        action_name: str = "gpt-4o",
    ) -> None:
        super().__init__()
        self.client = client
        self.subject = subject
        self.estimate_amount = estimate_amount
        self.action_kind = action_kind
        self.action_name = action_name
        # Track active reservations by run_id
        self._reservations: dict[str, str] = {}
        self._idempotency_keys: dict[str, str] = {}

    def on_llm_start(
        self,
        serialized: dict[str, Any],
        prompts: list[str],
        *,
        run_id: uuid.UUID,
        **kwargs: Any,
    ) -> None:
        """Create a budget reservation before each LLM call."""
        key = str(uuid.uuid4())
        self._idempotency_keys[str(run_id)] = key

        response = self.client.create_reservation(
            ReservationCreateRequest(
                idempotency_key=key,
                subject=self.subject,
                action=Action(kind=self.action_kind, name=self.action_name),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=self.estimate_amount),
                ttl_ms=60_000,
            )
        )

        if not response.is_success:
            error = response.get_error_response()
            if error and error.error == "BUDGET_EXCEEDED":
                raise BudgetExceededError(
                    error.message,
                    status=response.status,
                    error_code=error.error,
                    request_id=error.request_id,
                    details=error.details,
                )
            msg = error.message if error else (response.error_message or "Reservation failed")
            raise CyclesProtocolError(
                msg,
                status=response.status,
                error_code=error.error if error else None,
                request_id=error.request_id if error else None,
                details=error.details if error else None,
            )

        reservation_id = response.get_body_attribute("reservation_id")
        self._reservations[str(run_id)] = reservation_id

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: uuid.UUID,
        **kwargs: Any,
    ) -> None:
        """Commit actual cost after the LLM call completes."""
        run_key = str(run_id)
        reservation_id = self._reservations.pop(run_key, None)
        idempotency_key = self._idempotency_keys.pop(run_key, None)

        if not reservation_id or not idempotency_key:
            return

        # Extract token usage from LangChain's response
        token_usage = (response.llm_output or {}).get("token_usage", {})
        input_tokens = token_usage.get("prompt_tokens", 0)
        output_tokens = token_usage.get("completion_tokens", 0)

        actual_cost = (
            input_tokens * PRICE_PER_INPUT_TOKEN
            + output_tokens * PRICE_PER_OUTPUT_TOKEN
        )

        self.client.commit_reservation(
            reservation_id,
            CommitRequest(
                idempotency_key=f"commit-{idempotency_key}",
                actual=Amount(unit=Unit.USD_MICROCENTS, amount=actual_cost),
                metrics=CyclesMetrics(
                    tokens_input=input_tokens,
                    tokens_output=output_tokens,
                    model_version=token_usage.get("model_name", self.action_name),
                ),
            ),
        )

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: uuid.UUID,
        **kwargs: Any,
    ) -> None:
        """Release the reservation if the LLM call fails."""
        run_key = str(run_id)
        reservation_id = self._reservations.pop(run_key, None)
        idempotency_key = self._idempotency_keys.pop(run_key, None)

        if reservation_id and idempotency_key:
            self.client.release_reservation(
                reservation_id,
                ReleaseRequest(idempotency_key=f"release-{idempotency_key}"),
            )


# ---------------------------------------------------------------------------
# 2. Using the handler with a chat model
# ---------------------------------------------------------------------------
def simple_chain_example() -> None:
    """Run a simple LangChain invocation with budget protection."""
    config = CyclesConfig(
        base_url=os.environ.get("CYCLES_BASE_URL", "http://localhost:7878"),
        api_key=os.environ.get("CYCLES_API_KEY", "your-api-key"),
        tenant=os.environ.get("CYCLES_TENANT", "acme"),
    )
    client = CyclesClient(config)

    handler = CyclesBudgetHandler(
        client=client,
        subject=Subject(tenant=config.tenant, agent="langchain-agent"),
    )

    llm = ChatOpenAI(
        model="gpt-4o",
        callbacks=[handler],
    )

    print("=== Simple invocation ===")
    try:
        result = llm.invoke([HumanMessage(content="What is budget authority in one sentence?")])
        print(f"Response: {result.content}")
    except BudgetExceededError:
        print("Budget exhausted — cannot invoke LLM.")

    client.close()


# ---------------------------------------------------------------------------
# 3. Using with an agent that has tools
# ---------------------------------------------------------------------------
def agent_with_tools_example() -> None:
    """Run a LangChain agent with tools, each LLM call budget-guarded."""
    from langchain_core.tools import tool

    config = CyclesConfig(
        base_url=os.environ.get("CYCLES_BASE_URL", "http://localhost:7878"),
        api_key=os.environ.get("CYCLES_API_KEY", "your-api-key"),
        tenant=os.environ.get("CYCLES_TENANT", "acme"),
    )
    client = CyclesClient(config)

    handler = CyclesBudgetHandler(
        client=client,
        subject=Subject(tenant=config.tenant, agent="tool-agent", toolset="weather"),
    )

    @tool
    def get_weather(location: str) -> str:
        """Get the current weather for a location."""
        return f"72°F and sunny in {location}"

    llm = ChatOpenAI(model="gpt-4o", callbacks=[handler])
    llm_with_tools = llm.bind_tools([get_weather])

    print("\n=== Agent with tools ===")
    try:
        result = llm_with_tools.invoke(
            [HumanMessage(content="What's the weather in San Francisco?")]
        )
        print(f"Response: {result.content}")

        # If the model requested a tool call, show it
        if result.tool_calls:
            for tc in result.tool_calls:
                print(f"  Tool call: {tc['name']}({tc['args']})")
                tool_result = get_weather.invoke(tc["args"])
                print(f"  Tool result: {tool_result}")

    except BudgetExceededError:
        print("Budget exhausted — agent stopped.")

    client.close()


# ---------------------------------------------------------------------------
# 4. Run examples
# ---------------------------------------------------------------------------
def main() -> None:
    simple_chain_example()
    agent_with_tools_example()


if __name__ == "__main__":
    main()



"""runcycles - Python client for the Cycles budget-management protocol."""

from runcycles.client import AsyncCyclesClient, CyclesClient
from runcycles.config import CyclesConfig
from runcycles.context import CyclesContext, get_cycles_context
from runcycles.decorator import cycles, set_default_client, set_default_config
from runcycles.exceptions import (
    BudgetExceededError,
    CyclesError,
    CyclesProtocolError,
    CyclesTransportError,
    DebtOutstandingError,
    OverdraftLimitExceededError,
    ReservationExpiredError,
    ReservationFinalizedError,
)
from runcycles.models import (
    Action,
    Amount,
    Balance,
    BalanceResponse,
    Caps,
    CommitOveragePolicy,
    CommitRequest,
    CommitResponse,
    CommitStatus,
    CyclesMetrics,
    Decision,
    DecisionRequest,
    DecisionResponse,
    DryRunResult,
    ErrorCode,
    ErrorResponse,
    EventCreateRequest,
    EventCreateResponse,
    EventStatus,
    ExtendStatus,
    ReleaseRequest,
    ReleaseResponse,
    ReleaseStatus,
    ReservationCreateRequest,
    ReservationCreateResponse,
    ReservationDetail,
    ReservationExtendRequest,
    ReservationExtendResponse,
    ReservationListResponse,
    ReservationStatus,
    ReservationSummary,
    SignedAmount,
    Subject,
    Unit,
)
from runcycles.response import CyclesResponse

__all__ = [
    # Client
    "CyclesClient",
    "AsyncCyclesClient",
    # Config
    "CyclesConfig",
    # Decorator
    "cycles",
    "set_default_client",
    "set_default_config",
    # Context
    "CyclesContext",
    "get_cycles_context",
    # Response
    "CyclesResponse",
    # Exceptions
    "CyclesError",
    "CyclesProtocolError",
    "CyclesTransportError",
    "BudgetExceededError",
    "DebtOutstandingError",
    "OverdraftLimitExceededError",
    "ReservationExpiredError",
    "ReservationFinalizedError",
    # Models
    "Unit",
    "Amount",
    "SignedAmount",
    "Subject",
    "Action",
    "Caps",
    "Decision",
    "CyclesMetrics",
    "Balance",
    "CommitOveragePolicy",
    "ErrorCode",
    "ErrorResponse",
    "ReservationStatus",
    "CommitStatus",
    "ReleaseStatus",
    "ExtendStatus",
    "EventStatus",
    "ReservationCreateRequest",
    "CommitRequest",
    "ReleaseRequest",
    "ReservationExtendRequest",
    "DecisionRequest",
    "EventCreateRequest",
    "ReservationCreateResponse",
    "CommitResponse",
    "ReleaseResponse",
    "ReservationExtendResponse",
    "DecisionResponse",
    "EventCreateResponse",
    "DryRunResult",
    "ReservationDetail",
    "ReservationSummary",
    "ReservationListResponse",
    "BalanceResponse",
]

from importlib.metadata import version as _metadata_version

__version__ = _metadata_version("runcycles")



"""Exception hierarchy for the Cycles client."""

from __future__ import annotations

from typing import Any


class CyclesError(Exception):
    """Base exception for all Cycles client errors."""


class CyclesProtocolError(CyclesError):
    """Raised when the Cycles server returns a protocol-level error."""

    def __init__(
        self,
        message: str,
        *,
        status: int = 0,
        error_code: str | None = None,
        reason_code: str | None = None,
        retry_after_ms: int | None = None,
        request_id: str | None = None,
        details: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(message)
        self.status = status
        self.error_code = error_code
        self.reason_code = reason_code
        self.retry_after_ms = retry_after_ms
        self.request_id = request_id
        self.details = details

    def is_budget_exceeded(self) -> bool:
        return self.error_code == "BUDGET_EXCEEDED"

    def is_overdraft_limit_exceeded(self) -> bool:
        return self.error_code == "OVERDRAFT_LIMIT_EXCEEDED"

    def is_debt_outstanding(self) -> bool:
        return self.error_code == "DEBT_OUTSTANDING"

    def is_reservation_expired(self) -> bool:
        return self.error_code == "RESERVATION_EXPIRED"

    def is_reservation_finalized(self) -> bool:
        return self.error_code == "RESERVATION_FINALIZED"

    def is_idempotency_mismatch(self) -> bool:
        return self.error_code == "IDEMPOTENCY_MISMATCH"

    def is_unit_mismatch(self) -> bool:
        return self.error_code == "UNIT_MISMATCH"

    def is_retryable(self) -> bool:
        return self.error_code in ("INTERNAL_ERROR", "UNKNOWN") or self.status >= 500


class BudgetExceededError(CyclesProtocolError):
    """Raised when budget is insufficient for the reservation."""


class OverdraftLimitExceededError(CyclesProtocolError):
    """Raised when debt exceeds the overdraft limit."""


class DebtOutstandingError(CyclesProtocolError):
    """Raised when outstanding debt blocks new reservations."""


class ReservationExpiredError(CyclesProtocolError):
    """Raised when operating on an expired reservation."""


class ReservationFinalizedError(CyclesProtocolError):
    """Raised when operating on an already-finalized reservation."""


class CyclesTransportError(CyclesError):
    """Raised when a transport-level error occurs (connection timeout, DNS failure, etc.)."""

    def __init__(self, message: str, *, cause: Exception | None = None) -> None:
        super().__init__(message)
        self.cause = cause



"""Integrating Cycles with the Anthropic Python SDK.

Guards Anthropic Messages API calls with budget reservations and
demonstrates per-tool-call budget tracking for agentic workflows.

Requirements:
    pip install runcycles anthropic

Environment variables:
    CYCLES_BASE_URL    - Cycles server URL (default: http://localhost:7878)
    CYCLES_API_KEY     - Cycles API key
    CYCLES_TENANT      - Tenant identifier
    ANTHROPIC_API_KEY  - Anthropic API key
"""

import os
import uuid

from anthropic import Anthropic

from runcycles import (
    Action,
    Amount,
    BudgetExceededError,
    CommitRequest,
    CyclesClient,
    CyclesConfig,
    CyclesMetrics,
    ReservationCreateRequest,
    Subject,
    Unit,
    cycles,
    get_cycles_context,
    set_default_client,
)

# ---------------------------------------------------------------------------
# 1. Configure clients
# ---------------------------------------------------------------------------
config = CyclesConfig(
    base_url=os.environ.get("CYCLES_BASE_URL", "http://localhost:7878"),
    api_key=os.environ.get("CYCLES_API_KEY", "your-api-key"),
    tenant=os.environ.get("CYCLES_TENANT", "acme"),
    app="anthropic-example",
)
cycles_client = CyclesClient(config)
set_default_client(cycles_client)

anthropic_client = Anthropic()  # reads ANTHROPIC_API_KEY from env

# Pricing in USD microcents (1 USD = 100_000_000 microcents).
PRICE_PER_INPUT_TOKEN = 300       # $3.00 / 1M tokens
PRICE_PER_OUTPUT_TOKEN = 1_500    # $15.00 / 1M tokens


# ---------------------------------------------------------------------------
# 2. Simple decorator-based integration
# ---------------------------------------------------------------------------
@cycles(
    estimate=lambda prompt, **kw: (
        len(prompt.split()) * 2 * PRICE_PER_INPUT_TOKEN
        + kw.get("max_tokens", 1024) * PRICE_PER_OUTPUT_TOKEN
    ),
    actual=lambda result: (
        result["usage"]["input_tokens"] * PRICE_PER_INPUT_TOKEN
        + result["usage"]["output_tokens"] * PRICE_PER_OUTPUT_TOKEN
    ),
    action_kind="llm.completion",
    action_name="claude-sonnet-4-20250514",
    unit="USD_MICROCENTS",
    ttl_ms=60_000,
)
def send_message(prompt: str, max_tokens: int = 1024) -> dict:
    """Send a message to Claude with budget protection."""
    ctx = get_cycles_context()

    # Respect caps
    if ctx and ctx.has_caps() and ctx.caps.max_tokens:
        max_tokens = min(max_tokens, ctx.caps.max_tokens)

    response = anthropic_client.messages.create(
        model="claude-sonnet-4-20250514",
        max_tokens=max_tokens,
        messages=[{"role": "user", "content": prompt}],
    )

    if ctx:
        ctx.metrics = CyclesMetrics(
            tokens_input=response.usage.input_tokens,
            tokens_output=response.usage.output_tokens,
            model_version=response.model,
        )

    return {
        "content": response.content[0].text,
        "usage": {
            "input_tokens": response.usage.input_tokens,
            "output_tokens": response.usage.output_tokens,
        },
    }


# ---------------------------------------------------------------------------
# 3. Per-tool-call budget tracking (programmatic)
# ---------------------------------------------------------------------------
TOOLS = [
    {
        "name": "get_weather",
        "description": "Get current weather for a location.",
        "input_schema": {
            "type": "object",
            "properties": {"location": {"type": "string"}},
            "required": ["location"],
        },
    }
]


def handle_tool_call(tool_name: str, tool_input: dict) -> str:
    """Simulate executing a tool."""
    if tool_name == "get_weather":
        return f"72°F and sunny in {tool_input.get('location', 'unknown')}"
    return "Unknown tool"


def chat_with_tools(prompt: str) -> str:
    """Multi-turn conversation with tool use, each LLM call budget-guarded."""
    messages = [{"role": "user", "content": prompt}]
    turn = 0

    while True:
        turn += 1
        idempotency_key = f"tool-chat-{uuid.uuid4()}"

        # Reserve budget for this turn
        reserve_response = cycles_client.create_reservation(
            ReservationCreateRequest(
                idempotency_key=idempotency_key,
                subject=Subject(
                    tenant=config.tenant,
                    agent="tool-agent",
                    toolset="weather-tools",
                ),
                action=Action(
                    kind="llm.completion",
                    name="claude-sonnet-4-20250514",
                    tags=[f"turn-{turn}"],
                ),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=2_000_000),
                ttl_ms=30_000,
            )
        )

        if not reserve_response.is_success:
            error_resp = reserve_response.get_error_response()
            if error_resp and error_resp.error == "BUDGET_EXCEEDED":
                return "Budget exhausted — cannot continue conversation."
            return f"Reservation failed: {reserve_response.error_message}"

        reservation_id = reserve_response.get_body_attribute("reservation_id")

        # Call Claude
        response = anthropic_client.messages.create(
            model="claude-sonnet-4-20250514",
            max_tokens=1024,
            tools=TOOLS,
            messages=messages,
        )

        # Commit actual cost
        actual_cost = (
            response.usage.input_tokens * PRICE_PER_INPUT_TOKEN
            + response.usage.output_tokens * PRICE_PER_OUTPUT_TOKEN
        )
        cycles_client.commit_reservation(
            reservation_id,
            CommitRequest(
                idempotency_key=f"commit-{idempotency_key}",
                actual=Amount(unit=Unit.USD_MICROCENTS, amount=actual_cost),
                metrics=CyclesMetrics(
                    tokens_input=response.usage.input_tokens,
                    tokens_output=response.usage.output_tokens,
                    model_version=response.model,
                    custom={"turn": turn, "tool_use": response.stop_reason == "tool_use"},
                ),
            ),
        )

        # Process response
        if response.stop_reason == "end_turn":
            # Final text response
            for block in response.content:
                if block.type == "text":
                    return block.text
            return ""

        if response.stop_reason == "tool_use":
            # Extract tool calls and execute them
            messages.append({"role": "assistant", "content": response.content})
            tool_results = []
            for block in response.content:
                if block.type == "tool_use":
                    result = handle_tool_call(block.name, block.input)
                    tool_results.append({
                        "type": "tool_result",
                        "tool_use_id": block.id,
                        "content": result,
                    })
            messages.append({"role": "user", "content": tool_results})
        else:
            # Unexpected stop reason
            return f"Stopped with reason: {response.stop_reason}"

        if turn >= 5:
            return "Maximum turns reached."


# ---------------------------------------------------------------------------
# 4. Run it
# ---------------------------------------------------------------------------
def main() -> None:
    print("=== Simple message ===")
    try:
        result = send_message("What is budget authority in one sentence?")
        print(f"Response: {result['content']}")
        print(f"Tokens: {result['usage']}")
    except BudgetExceededError:
        print("Budget exhausted.")

    print("\n=== Tool-use conversation ===")
    answer = chat_with_tools("What's the weather like in San Francisco?")
    print(f"Answer: {answer}")


if __name__ == "__main__":
    main()



"""Tests for Pydantic models."""

import pytest
from pydantic import ValidationError

from runcycles.models import (
    Action,
    Amount,
    Balance,
    Caps,
    CommitOveragePolicy,
    CommitRequest,
    CyclesMetrics,
    Decision,
    DecisionResponse,
    DryRunResult,
    ErrorCode,
    ErrorResponse,
    ReleaseRequest,
    ReservationCreateRequest,
    ReservationCreateResponse,
    ReservationDetail,
    ReservationExtendRequest,
    ReservationStatus,
    SignedAmount,
    Subject,
    Unit,
)


class TestAmount:
    def test_create(self) -> None:
        a = Amount(unit=Unit.USD_MICROCENTS, amount=1000)
        assert a.unit == Unit.USD_MICROCENTS
        assert a.amount == 1000

    def test_serialize(self) -> None:
        a = Amount(unit=Unit.USD_MICROCENTS, amount=500)
        d = a.model_dump()
        assert d == {"unit": "USD_MICROCENTS", "amount": 500}

    def test_deserialize(self) -> None:
        a = Amount.model_validate({"unit": "TOKENS", "amount": 42})
        assert a.unit == Unit.TOKENS
        assert a.amount == 42


class TestSubject:
    def test_standard_field_check(self) -> None:
        s = Subject(tenant="acme")
        assert s.has_at_least_one_standard_field()

    def test_no_standard_field(self) -> None:
        s = Subject(dimensions={"foo": "bar"})
        assert not s.has_at_least_one_standard_field()

    def test_serialize_excludes_none(self) -> None:
        s = Subject(tenant="acme", agent="bot")
        d = s.model_dump(exclude_none=True)
        assert d == {"tenant": "acme", "agent": "bot"}
        assert "workspace" not in d


class TestCaps:
    def test_tool_allowed_with_allowlist(self) -> None:
        c = Caps(tool_allowlist=["search", "calc"])
        assert c.is_tool_allowed("search")
        assert not c.is_tool_allowed("code_exec")

    def test_tool_denied_with_denylist(self) -> None:
        c = Caps(tool_denylist=["code_exec"])
        assert c.is_tool_allowed("search")
        assert not c.is_tool_allowed("code_exec")

    def test_tool_allowed_no_lists(self) -> None:
        c = Caps()
        assert c.is_tool_allowed("anything")


class TestCyclesMetrics:
    def test_is_empty(self) -> None:
        m = CyclesMetrics()
        assert m.is_empty()

    def test_not_empty(self) -> None:
        m = CyclesMetrics(tokens_input=100)
        assert not m.is_empty()

    def test_put_custom(self) -> None:
        m = CyclesMetrics()
        m.put_custom("cache_hit", True)
        assert m.custom == {"cache_hit": True}
        assert not m.is_empty()


class TestReservationCreateRequest:
    def test_serialize(self) -> None:
        r = ReservationCreateRequest(
            idempotency_key="req-001",
            subject=Subject(tenant="acme"),
            action=Action(kind="llm.completion", name="gpt-4"),
            estimate=Amount(unit=Unit.USD_MICROCENTS, amount=1000),
            ttl_ms=30000,
            overage_policy=CommitOveragePolicy.REJECT,
        )
        d = r.model_dump(exclude_none=True)
        assert d["idempotency_key"] == "req-001"
        assert d["subject"]["tenant"] == "acme"
        assert d["estimate"]["amount"] == 1000
        assert d["ttl_ms"] == 30000


class TestCommitRequest:
    def test_serialize(self) -> None:
        r = CommitRequest(
            idempotency_key="commit-001",
            actual=Amount(unit=Unit.USD_MICROCENTS, amount=800),
            metrics=CyclesMetrics(tokens_input=100, tokens_output=50),
        )
        d = r.model_dump(exclude_none=True)
        assert d["actual"]["amount"] == 800
        assert d["metrics"]["tokens_input"] == 100


class TestReservationCreateResponse:
    def test_from_server_response(self) -> None:
        data = {
            "decision": "ALLOW",
            "reservation_id": "res_abc-123",
            "expires_at_ms": 1710000060000,
            "affected_scopes": ["tenant:acme"],
            "scope_path": "tenant:acme",
            "reserved": {"unit": "USD_MICROCENTS", "amount": 500000},
            "caps": None,
            "reason_code": None,
            "retry_after_ms": None,
        }
        r = ReservationCreateResponse.model_validate(data)
        assert r.decision == Decision.ALLOW
        assert r.reservation_id == "res_abc-123"
        assert r.is_allowed()
        assert not r.is_denied()
        assert r.reserved is not None
        assert r.reserved.amount == 500000

    def test_deny(self) -> None:
        data = {"decision": "DENY", "reason_code": "BUDGET_EXCEEDED", "affected_scopes": ["tenant:acme"]}
        r = ReservationCreateResponse.model_validate(data)
        assert r.is_denied()
        assert not r.is_allowed()

    def test_allow_with_caps(self) -> None:
        data = {
            "decision": "ALLOW_WITH_CAPS",
            "reservation_id": "res_123",
            "affected_scopes": ["tenant:acme"],
            "caps": {"max_tokens": 500, "tool_denylist": ["code_exec"]},
        }
        r = ReservationCreateResponse.model_validate(data)
        assert r.decision == Decision.ALLOW_WITH_CAPS
        assert r.caps is not None
        assert r.caps.max_tokens == 500


class TestErrorResponse:
    def test_parse(self) -> None:
        data = {
            "error": "BUDGET_EXCEEDED",
            "message": "Insufficient budget",
            "request_id": "req-123",
        }
        e = ErrorResponse.model_validate(data)
        assert e.error_code == ErrorCode.BUDGET_EXCEEDED
        assert e.message == "Insufficient budget"

    def test_unknown_error(self) -> None:
        data = {"error": "SOME_FUTURE_ERROR", "message": "Something new", "request_id": "req-456"}
        e = ErrorResponse.model_validate(data)
        assert e.error_code == ErrorCode.UNKNOWN


class TestErrorCode:
    def test_retryable(self) -> None:
        assert ErrorCode.INTERNAL_ERROR.is_retryable
        assert ErrorCode.UNKNOWN.is_retryable
        assert not ErrorCode.BUDGET_EXCEEDED.is_retryable

    def test_from_string(self) -> None:
        assert ErrorCode.from_string("BUDGET_EXCEEDED") == ErrorCode.BUDGET_EXCEEDED
        assert ErrorCode.from_string("NONSENSE") == ErrorCode.UNKNOWN
        assert ErrorCode.from_string(None) is None


class TestAmountValidation:
    def test_rejects_negative_amount(self) -> None:
        with pytest.raises(ValidationError):
            Amount(unit=Unit.USD_MICROCENTS, amount=-1)

    def test_allows_zero_amount(self) -> None:
        a = Amount(unit=Unit.USD_MICROCENTS, amount=0)
        assert a.amount == 0

    def test_signed_amount_allows_negative(self) -> None:
        s = SignedAmount(unit=Unit.USD_MICROCENTS, amount=-500)
        assert s.amount == -500


class TestCapsToolPrecedence:
    def test_allowlist_takes_precedence_over_denylist(self) -> None:
        """Spec: If tool_allowlist is non-empty, ONLY those tools are allowed (denylist ignored)."""
        c = Caps(tool_allowlist=["search", "calc"], tool_denylist=["search"])
        # search is on both lists — allowlist takes precedence
        assert c.is_tool_allowed("search")
        assert not c.is_tool_allowed("code_exec")

    def test_empty_allowlist_blocks_all(self) -> None:
        """Empty allowlist means no tools are allowed."""
        c = Caps(tool_allowlist=[])
        assert not c.is_tool_allowed("anything")


class TestDryRunResult:
    def test_is_allowed(self) -> None:
        r = DryRunResult(decision=Decision.ALLOW)
        assert r.is_allowed()
        assert not r.is_denied()

    def test_is_denied(self) -> None:
        r = DryRunResult(decision=Decision.DENY, reason_code="BUDGET_EXCEEDED")
        assert r.is_denied()
        assert not r.is_allowed()

    def test_has_caps(self) -> None:
        r = DryRunResult(decision=Decision.ALLOW_WITH_CAPS, caps=Caps(max_tokens=100))
        assert r.has_caps()
        assert r.is_allowed()


def _make_detail(status: ReservationStatus) -> ReservationDetail:
    """Helper to build a ReservationDetail with all required fields."""
    return ReservationDetail(
        reservation_id="rsv_1",
        status=status,
        subject=Subject(tenant="acme"),
        action=Action(kind="test", name="test"),
        reserved=Amount(unit=Unit.USD_MICROCENTS, amount=100),
        created_at_ms=1000000,
        expires_at_ms=2000000,
        scope_path="tenant:acme",
        affected_scopes=["tenant:acme"],
    )


class TestReservationDetail:
    def test_is_active(self) -> None:
        r = _make_detail(ReservationStatus.ACTIVE)
        assert r.is_active()
        assert not r.is_committed()
        assert not r.is_released()
        assert not r.is_expired()

    def test_is_committed(self) -> None:
        r = _make_detail(ReservationStatus.COMMITTED)
        assert r.is_committed()

    def test_is_released(self) -> None:
        r = _make_detail(ReservationStatus.RELEASED)
        assert r.is_released()

    def test_is_expired(self) -> None:
        r = _make_detail(ReservationStatus.EXPIRED)
        assert r.is_expired()


class TestDecisionResponse:
    def test_allow(self) -> None:
        r = DecisionResponse(decision=Decision.ALLOW)
        assert r.is_allowed()
        assert not r.is_denied()

    def test_deny(self) -> None:
        r = DecisionResponse(decision=Decision.DENY, reason_code="BUDGET_EXCEEDED")
        assert r.is_denied()
        assert not r.is_allowed()

    def test_allow_with_caps(self) -> None:
        r = DecisionResponse(decision=Decision.ALLOW_WITH_CAPS, caps=Caps(max_tokens=500))
        assert r.is_allowed()


class TestRequiredFields:
    """Validate spec-mandated required fields are enforced."""

    def test_reservation_result_requires_decision(self) -> None:
        with pytest.raises(ValidationError):
            ReservationCreateResponse.model_validate({"affected_scopes": ["tenant:acme"]})

    def test_reservation_result_requires_affected_scopes(self) -> None:
        with pytest.raises(ValidationError):
            ReservationCreateResponse.model_validate({"decision": "ALLOW"})

    def test_commit_result_requires_status_and_charged(self) -> None:
        from runcycles.models import CommitResponse
        with pytest.raises(ValidationError):
            CommitResponse.model_validate({"status": "COMMITTED"})
        with pytest.raises(ValidationError):
            CommitResponse.model_validate({"charged": {"unit": "USD_MICROCENTS", "amount": 100}})

    def test_release_result_requires_status_and_released(self) -> None:
        from runcycles.models import ReleaseResponse
        with pytest.raises(ValidationError):
            ReleaseResponse.model_validate({"status": "RELEASED"})
        with pytest.raises(ValidationError):
            ReleaseResponse.model_validate({"released": {"unit": "USD_MICROCENTS", "amount": 100}})

    def test_extend_result_requires_status_and_expires(self) -> None:
        from runcycles.models import ReservationExtendResponse
        with pytest.raises(ValidationError):
            ReservationExtendResponse.model_validate({"status": "ACTIVE"})
        with pytest.raises(ValidationError):
            ReservationExtendResponse.model_validate({"expires_at_ms": 9999999999})

    def test_event_result_requires_status_and_event_id(self) -> None:
        from runcycles.models import EventCreateResponse
        with pytest.raises(ValidationError):
            EventCreateResponse.model_validate({"status": "APPLIED"})
        with pytest.raises(ValidationError):
            EventCreateResponse.model_validate({"event_id": "evt_123"})

    def test_balance_requires_scope_scope_path_remaining(self) -> None:
        with pytest.raises(ValidationError):
            Balance.model_validate({"scope_path": "/", "remaining": {"unit": "USD_MICROCENTS", "amount": 100}})
        with pytest.raises(ValidationError):
            Balance.model_validate({"scope": "tenant:acme", "remaining": {"unit": "USD_MICROCENTS", "amount": 100}})
        with pytest.raises(ValidationError):
            Balance.model_validate({"scope": "tenant:acme", "scope_path": "/"})

    def test_decision_result_requires_decision(self) -> None:
        with pytest.raises(ValidationError):
            DecisionResponse.model_validate({})


class TestFieldConstraints:
    """Validate spec-mandated field constraints are enforced."""

    def test_subject_tenant_max_length(self) -> None:
        with pytest.raises(ValidationError):
            Subject(tenant="x" * 129)

    def test_subject_tenant_at_max_length(self) -> None:
        s = Subject(tenant="x" * 128)
        assert len(s.tenant) == 128

    def test_subject_dimensions_max_entries(self) -> None:
        dims = {f"k{i}": "v" for i in range(17)}
        with pytest.raises(ValidationError):
            Subject(tenant="acme", dimensions=dims)

    def test_subject_dimension_value_max_length(self) -> None:
        with pytest.raises(ValidationError):
            Subject(tenant="acme", dimensions={"key": "v" * 257})

    def test_action_kind_max_length(self) -> None:
        with pytest.raises(ValidationError):
            Action(kind="x" * 65, name="ok")

    def test_action_name_max_length(self) -> None:
        with pytest.raises(ValidationError):
            Action(kind="ok", name="x" * 257)

    def test_action_tags_max_items(self) -> None:
        with pytest.raises(ValidationError):
            Action(kind="ok", name="ok", tags=[f"t{i}" for i in range(11)])

    def test_action_tag_item_max_length(self) -> None:
        with pytest.raises(ValidationError):
            Action(kind="ok", name="ok", tags=["x" * 65])

    def test_caps_max_tokens_non_negative(self) -> None:
        with pytest.raises(ValidationError):
            Caps(max_tokens=-1)

    def test_caps_max_steps_non_negative(self) -> None:
        with pytest.raises(ValidationError):
            Caps(max_steps_remaining=-1)

    def test_caps_cooldown_non_negative(self) -> None:
        with pytest.raises(ValidationError):
            Caps(cooldown_ms=-1)

    def test_metrics_tokens_input_non_negative(self) -> None:
        with pytest.raises(ValidationError):
            CyclesMetrics(tokens_input=-1)

    def test_metrics_tokens_output_non_negative(self) -> None:
        with pytest.raises(ValidationError):
            CyclesMetrics(tokens_output=-1)

    def test_metrics_latency_non_negative(self) -> None:
        with pytest.raises(ValidationError):
            CyclesMetrics(latency_ms=-1)

    def test_metrics_model_version_max_length(self) -> None:
        with pytest.raises(ValidationError):
            CyclesMetrics(model_version="x" * 129)

    def test_idempotency_key_min_length(self) -> None:
        with pytest.raises(ValidationError):
            ReservationCreateRequest(
                idempotency_key="",
                subject=Subject(tenant="acme"),
                action=Action(kind="test", name="test"),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=100),
            )

    def test_idempotency_key_max_length(self) -> None:
        with pytest.raises(ValidationError):
            ReservationCreateRequest(
                idempotency_key="x" * 257,
                subject=Subject(tenant="acme"),
                action=Action(kind="test", name="test"),
                estimate=Amount(unit=Unit.USD_MICROCENTS, amount=100),
            )

    def test_extend_by_ms_min(self) -> None:
        with pytest.raises(ValidationError):
            ReservationExtendRequest(idempotency_key="key-1", extend_by_ms=0)

    def test_extend_by_ms_max(self) -> None:
        with pytest.raises(ValidationError):
            ReservationExtendRequest(idempotency_key="key-1", extend_by_ms=86_400_001)

    def test_release_reason_max_length(self) -> None:
        with pytest.raises(ValidationError):
            ReleaseRequest(idempotency_key="key-1", reason="x" * 257)

    def test_reservation_detail_requires_all_spec_fields(self) -> None:
        """ReservationDetail spec requires many fields."""
        with pytest.raises(ValidationError):
            ReservationDetail(reservation_id="rsv_1", status=ReservationStatus.ACTIVE)

    def test_reservation_list_requires_reservations(self) -> None:
        from runcycles.models import ReservationListResponse
        with pytest.raises(ValidationError):
            ReservationListResponse.model_validate({})

    def test_balance_query_requires_balances(self) -> None:
        from runcycles.models import BalanceResponse
        with pytest.raises(ValidationError):
            BalanceResponse.model_validate({})
